@@ -58,54 +58,11 @@ use Data::Dumper;
 use XML::LibXML::PrettyPrint;
 
 use Geo::OGC::Service;
+use Geo::OGC::Service::Filter ':all';
 use vars qw(@ISA);
-push @ISA, qw(Geo::OGC::Service::Common);
+push @ISA, qw(Geo::OGC::Service::Filter);
 
-our $VERSION = '0.09';
-
-our %gml_geometry_type = (
-    Envelope => 1,
-    Point => 1,
-    MultiPoint => 1,
-    LineString => 1,
-    MultiLineString => 1,
-    Polygon => 1,
-    MultiPolygon => 1,
-    MultiGeometry => 1,
-    );
-
-our %spatial2op = ( 
-    Disjoint => 1,
-    Equals => 1,
-    Touches => 1,
-    Within => 1,
-    Overlaps => 1,
-    Crosses => 1,
-    Intersects => 1,
-    Contains => 1
-    ); # there are more in PostGIS
-
-our %temporal_operators = (
-    After => 1,
-    Before => 1,
-    Begins => 1,
-    BegunBy => 1,
-    TContains => 1,
-    During => 1,
-    TEquals => 1,
-    TOverlaps => 1,
-    Meets => 1,
-    OverlappedBy => 1,
-    MetBy => 1,
-    EndedBy => 1
-    );
-
-my %functions = (
-    abs => ['xs:int', [[int => 'xs:int']]],
-    acos => ['xs:double', [[value => 'xs:double']]],
-    Area => ['xs:double', [[geometry => 'gml:AbstractGeometryType']]],
-    within => ['xs:boolean', [[geometry => 'gml:AbstractGeometryType'], [geometry => 'gml:AbstractGeometryType']]]
-    ); # lots missing
+our $VERSION = '0.10';
 
 # GDAL and PostgreSQL data type to XML data type
 our %type_map = (
@@ -390,107 +347,6 @@ sub FeatureTypeList  {
     $writer->close_element;
 }
 
-sub Filter_Capabilities  {
-    my ($self, $writer) = @_;
-    my $ns = $self->{version} eq '2.0.0' ? 'fes' : 'ogc';
-    $writer->open_element($ns.':Filter_Capabilities');
-
-    # Conformance
-    my %Constraints = ( 
-        ImplementsQuery => [['ows:NoValues'], ['ows:DefaultValue' => 'TRUE']],
-        ImplementsAdHocQuery => [['ows:NoValues'], ['ows:DefaultValue' => 'TRUE']],
-        ImplementsFunctions => [['ows:NoValues'], ['ows:DefaultValue' => 'TRUE']],
-        ImplementsMinStandardFilter => [['ows:NoValues'], ['ows:DefaultValue' => 'TRUE']],
-        ImplementsStandardFilter => [['ows:NoValues'], ['ows:DefaultValue' => 'FALSE']],
-        ImplementsMinSpatialFilter => [['ows:NoValues'], ['ows:DefaultValue' => 'TRUE']],
-        ImplementsSpatialFilter => [['ows:NoValues'], ['ows:DefaultValue' => 'FALSE']],
-        ImplementsMinTemporalFilter => [['ows:NoValues'], ['ows:DefaultValue' => 'TRUE']],
-        ImplementsTemporalFilter => [['ows:NoValues'], ['ows:DefaultValue' => 'TRUE']],
-        ImplementsVersionNav => [['ows:NoValues'], ['ows:DefaultValue' => 'FALSE']],
-        ImplementsSorting => [['ows:AllowedValues' => [['ows:Value' => 'ASC'], ['ows:Value' => 'DESC']]], 
-                              ['ows:DefaultValue' => 'ASC']],
-        ImplementsExtendedOperators => [['ows:NoValues'], ['ows:DefaultValue' => 'FALSE']],
-        );
-    my @c;
-    for my $key (keys %Constraints) {
-        push @c, [$ns.':Constraint', {name=>$key}, $Constraints{$key}];
-    }
-    $writer->element($ns.':Conformance', \@c);
-
-    my @ids;
-    if ($ns eq 'ogc') {
-        @ids = ([$ns.':FID']);
-    } else {
-        @ids = (['fes:ResourceIdentifier', {name => 'fes:ResourceId'}]);
-    }
-    $writer->element($ns.':Id_Capabilities', \@ids);
-
-    my @operators = ();    
-    for my $o (qw/LessThan GreaterThan LessThanOrEqualTo GreaterThanOrEqualTo EqualTo NotEqualTo Like Between Null/) {
-        if ($ns eq 'ogc') {
-            push @operators, [$ns.':ComparisonOperator', 'PropertyIs'.$o];
-        } else {
-            push @operators, [$ns.':ComparisonOperator', { name => 'PropertyIs'.$o}];
-        }
-    }
-    $writer->element($ns.':Scalar_Capabilities', 
-                [[$ns.':LogicalOperators'], # empty ?
-                 [$ns.':ComparisonOperators', \@operators]]);
-
-    my @operands = ();
-    for my $o (keys %gml_geometry_type) {
-        if ($ns eq 'ogc') {
-            push @operands, [$ns.':GeometryOperand', 'gml:'.$o];
-        } else {
-            push @operands, [$ns.':GeometryOperand', { name => 'gml:'.$o }];
-        }
-    }
-    @operators = ();
-    my @op = keys %spatial2op;
-    push @op, (qw/DWithin BBOX/);
-    for my $o (@op) {
-        push @operators, [$ns.':SpatialOperator', { name => $o }];
-    }
-    $writer->element($ns.':Spatial_Capabilities', 
-                [[$ns.':GeometryOperands', \@operands],
-                 [$ns.':SpatialOperators', \@operators]]);
-
-    # Temporal_Capabilities
-
-    @operands = ();
-    for my $o (qw/TimeInstant TimePeriod/) {
-        if ($ns eq 'ogc') {
-            push @operands, [$ns.':GeometryOperand', 'gml:'.$o];
-        } else {
-            push @operands, [$ns.':GeometryOperand', { name => 'gml:'.$o }];
-        }
-    }
-    @operators = ();
-    @op = keys %temporal_operators;
-    for my $o (@op) {
-        push @operators, [$ns.':TemporalOperator', { name => $o }];
-    }
-    $writer->element($ns.':Temporal_Capabilities', 
-                     [[$ns.':TemporalOperands', \@operands],
-                      [$ns.':TemporalOperators', \@operators]]);
-
-    # Functions
-    
-    my @functions;
-    for my $f (sort keys %functions) {
-        my @args = ();
-        my $args = $functions{$f}[1];
-        for my $arg (@$args) {
-            push @args, [$ns.':Argument', { name => $arg->[0]}, [$ns.':Type' => $arg->[1]]];
-        }
-        push @functions, [$ns.':Function', { name => $f }, 
-                          [[$ns.':Returns' => $functions{$f}[0]] ,[$ns.':Arguments' => \@args]]];
-    }
-    $writer->element($ns.':Functions', \@functions);
-    
-    $writer->close_element;
-}
-
 =pod
 
 =head3 DescribeFeatureType
@@ -606,22 +462,24 @@ with layer name falling back to FeatureType).
 
 sub GetFeature {
     my ($self) = @_;
-    
+
     my $query = $self->{request}{queries}[0]; # actually we should loop through all queries?
     my ($typename) = split(/\s*,\s*/, $query->{typename});
     
     unless ($typename) {
         $self->error({ exceptionCode => 'MissingParameterValue',
-                       locator => 'typeName' });
+                       locator => 'typeName' },
+                     [$self->CORS]);
         return;
     }
-    
+
     my $type = $self->feature_type($typename);
 
     unless ($type) {
         $self->error({ exceptionCode => 'InvalidParameterValue',
                        locator => 'typeName',
-                       ExceptionText => "Type '$typename' is not available" });
+                       ExceptionText => "Type '$typename' is not available" },
+                     [$self->CORS]);
         return;
     }
 
@@ -635,7 +493,8 @@ sub GetFeature {
     elsif ($type && $type->{Table}) {
 
         my $filter = filter2sql($query->{filter}, $type) // '';
-        my ($epsg) = $query->{EPSG} =~ /(\d+)/;
+        my $epsg;
+        ($epsg) = $query->{EPSG} =~ /(\d\d\d\d+)/ if $query->{EPSG};
 
         # pseudo_credentials: these fields are required to be in the filter and they are not included as attributes
         my ($pseudo_credentials, @pseudo_credentials) = pseudo_credentials($type);
@@ -666,7 +525,7 @@ sub GetFeature {
             next if $type->{Schema}{$col}{out_type} eq 'gml:GeometryPropertyType' && not $is_geometry;
 
             if (defined $epsg and $is_geometry) {
-                push @cols, "st_transform(\"$col\",$epsg) as \"$name\"";
+                push @cols, "ST_Transform(\"$col\",$epsg) as \"$name\"";
             } else {
                 $filter =~ s/$name/$col/g if $filter;
                 $name = 'gml_id' if defined $type->{"gml:id"} && $col eq $type->{"gml:id"};
@@ -676,12 +535,12 @@ sub GetFeature {
 
         my $geom = '"'.$type->{GeometryColumn}.'"';
         
-        my $sql = "select ".join(',',@cols)." from \"$type->{Table}\" where ST_IsValid($geom)";
+        my $sql = "SELECT ".join(',',@cols)." FROM \"$type->{Table}\" WHERE ST_IsValid($geom)";
 
-        $geom = "st_transform($geom,$epsg)" if defined $epsg;
+        $geom = "ST_Transform($geom,$epsg)" if defined $epsg;
         $filter =~ s/GeometryColumn/$type->{GeometryColumn}/g if $filter;
         $sql .= " AND $filter" if $filter;
-        $sql .= " AND ($geom && ST_MakeEnvelope(".join(",", @{$query->{BBOX}})."))" if $query->{BBOX};
+        $sql .= " AND ($geom && ST_Transform(ST_MakeEnvelope(".join(",", @{$query->{BBOX}}).")),$type->{SRID})" if $query->{BBOX};
 
         print STDERR "$sql\n" if $self->{debug};
         eval {
@@ -693,7 +552,8 @@ sub GetFeature {
     unless ($layer) {
         $self->error({ exceptionCode => 'InvalidParameterValue',
                        locator => 'typeName',
-                       ExceptionText => "Type '$typename' is not available (or there was an internal or configuration error)" });
+                       ExceptionText => "Type '$typename' is not available (or there was an internal or configuration error)" },
+                     [$self->CORS]);
         return;
     }
 
@@ -709,6 +569,7 @@ sub GetFeature {
 
     my $i = 0;
     my $count = $query->{count};
+
     if (Geo::GDAL::VersionInfo >= 2010000) {
         # use streaming object
         #say STDERR "use streaming object";
@@ -1078,9 +939,11 @@ sub feature_type {
                 my $out_name = $in_name;
 
                 # field name adjustments as GDAL does them
-                $out_name =~ s/ /_/g; 
+                $out_name =~ s/ /_/g;
+                $out_name =~ s/-/_/g;
                 
                 # extra name adjustments, needed by QGIS
+                # this has no effect now since no 'use utf8'.. FIXME?
                 $out_name =~ s/[åö]/o/g;
                 $out_name =~ s/ä/a/g;
                 $out_name =~ s/[ÅÖ]/O/g;
@@ -1101,12 +964,12 @@ sub feature_type {
 sub feature_types_in_data_source {
     my ($self, $type) = @_;
     return unless $type->{DataSource} =~ /^Pg:/;
-
+    
     my $dbi = DataSource2dbi($type->{DataSource});
     my ($connect, $user, $pass) = split / /, $dbi;
     my $dbh = DBI->connect($connect, $user, $pass, { PrintError => 0, RaiseError => 0 }) 
         or croak("Can't connect to database '$connect': ".$DBI::errstr);
-    $dbh->{pg_enable_utf8} = 1;
+    #$dbh->{pg_enable_utf8} = 1; Use of the default is highly encouraged.
     my $sth = $dbh->table_info( '', 'public', undef, "'TABLE','VIEW'" );
     my @tables;
     while (my $data = $sth->fetchrow_hashref) {
@@ -1159,12 +1022,18 @@ sub feature_types_in_data_source {
 
             my $prefix = $type->{prefix};
             my $shortname = $table.'.'.$geom;
-            # wash the name
+
+            # XML layer names can't have spaces
             $shortname =~ s/ /_/g;
-            $shortname =~ s/[åö]/o/g;
-            $shortname =~ s/ä/a/g;
-            $shortname =~ s/[ÅÖ]/O/g;
-            $shortname =~ s/Ä/A/g;
+                
+            if (0) {
+                # if this is enabled then use utf8 is needed at least for åäöÅÄÖ
+                $shortname =~ s/[åö]/o/g;
+                $shortname =~ s/ä/a/g;
+                $shortname =~ s/[ÅÖ]/O/g;
+                $shortname =~ s/Ä/A/g;
+            }
+
             my $name = $prefix.'.'.$shortname;
             if ($type->{allow} and !($type->{allow}{$shortname} or $type->{allow}{$name})) {
                 print STDERR "not allowed.\n" if $self->{debug} > 2;
@@ -1182,6 +1051,7 @@ sub feature_types_in_data_source {
                 DefaultSRS => "$auth_name:$auth_srid",
                 Table => $table,
                 GeometryColumn => $geom,
+                SRID => $auth_srid,
                 Schema => \%schema
             };
             for my $n ($shortname, $name) {
@@ -1399,251 +1269,6 @@ sub transaction_sql {
     return \%dbisql;
 }
 
-# convert OGC Filter XML to SQL
-sub filter2sql {
-    my ($node, $type) = @_;
-    if (!$node) {
-        return "";
-    }
-    my ($ns, $name) = parse_tag($node);
-
-    if ($name eq 'Literal') {
-        return "'".$node->firstChild->data."'";
-
-    } elsif ($name eq 'PropertyName') {
-        my $ref = strip($node->textContent);
-        $ref = $type->{GeometryColumn} if $ref eq 'geometryProperty';
-        return '"'.$ref.'"';
-
-    } elsif ($name eq 'ValueReference') {
-        my $ref = strip($node->textContent);
-        $ref =~ s/^\w+://;
-        $ref = $type->{GeometryColumn} if $ref eq 'geometryProperty';
-        return '"'.$ref.'"';
-
-    } elsif ($name eq 'ResourceId') {
-        my $id = $node->getAttribute('rid');
-        return $type->{"gml:id"}." = '$id'"; # $type->{"gml:id"} may not be set!
-
-    } elsif ($name eq 'FeatureId') {
-        my $id = $node->getAttribute('fid');
-        return $type->{"gml:id"}." = '$id'"; # $type->{"gml:id"} may not be set!
-
-    } elsif ($name eq 'GmlObjectId') {
-        my $id = $node->getAttribute('gml:id');
-        return $type->{"gml:id"}." = '$id'"; # $type->{"gml:id"} may not be set!
-
-    } elsif ($name eq 'PropertyIsEqualTo') {
-        $node = $node->firstChild;
-        return '('.filter2sql($node, $type).' = '.filter2sql($node->nextSibling, $type).')';
-
-    } elsif ($name eq 'PropertyIsNotEqualTo') {
-        $node = $node->firstChild;
-        return '('.filter2sql($node, $type).' != '.filter2sql($node->nextSibling, $type).')';
-
-    } elsif ($name eq 'PropertyIsLessThan') {
-        $node = $node->firstChild;
-        return '('.filter2sql($node, $type).' < '.filter2sql($node->nextSibling, $type).')';
-
-    } elsif ($name eq 'PropertyIsGreaterThan') {
-        $node = $node->firstChild;
-        return '('.filter2sql($node, $type).' > '.filter2sql($node->nextSibling, $type).')';
-
-    } elsif ($name eq 'PropertyIsLessThanOrEqualTo') {
-        $node = $node->firstChild;
-        return '('.filter2sql($node, $type).' <= '.filter2sql($node->nextSibling, $type).')';
-
-    } elsif ($name eq 'PropertyIsGreaterThanOrEqualTo') {
-        $node = $node->firstChild;
-        return '('.filter2sql($node, $type).' >= '.filter2sql($node->nextSibling, $type).')';
-
-    } elsif ($name eq 'PropertyIsBetween') {
-        $node = $node->firstChild;
-        my $property = filter2sql($node->firstChild, $type);
-        $node = $node->nextSibling;
-        return '('.$property.' >= '.filter2sql($node, $type).' AND '.$property.'<='.filter2sql($node->nextSibling).')';
-
-    } elsif ($name eq 'PropertyIsLike') {
-        $node = $node->firstChild;
-        return '('.filter2sql($node, $type).' ~ '.filter2sql($node->nextSibling, $type).')';
-
-    } elsif ($name eq 'PropertyIsNull') {
-        return '('.filter2sql($node->firstChild, $type).' ISNULL)';
-
-    } elsif ($name eq 'And') {
-        $node = $node->firstChild;
-        my $p = '('.filter2sql($node, $type);
-        while ($node = $node->nextSibling) {
-            $p .= ' AND '.filter2sql($node, $type);
-        }
-        return $p.')';
-
-    } elsif ($name eq 'Or') {
-        $node = $node->firstChild;
-        my $p = '('.filter2sql($node, $type);
-        while ($node = $node->nextSibling) {
-            $p .= ' OR '.filter2sql($node, $type);
-        }
-        return $p.')';
-    } elsif ($name eq 'Not') {
-        return '(NOT '.filter2sql($node->firstChild, $type).')';
-
-    } elsif ($ns eq 'gml' and $gml_geometry_type{$name}) {
-        return node2sql($node);
-
-    } elsif ($name eq 'BBOX') {
-        $node = $node->firstChild;
-        my ($ns, $n) = parse_tag($node);
-        if ($n eq 'Envelope') {
-            my $envelope = filter2sql($node, $type);
-            return "(GeometryColumn && $envelope)";
-        } else {
-            my $property = filter2sql($node, $type);
-            $node = $node->nextSibling;
-            my $envelope = filter2sql($node, $type);
-            return "($property && $envelope)";
-        }
-
-    } elsif ($spatial2op{$name}) {
-        $node = $node->firstChild;
-        my $geom1 = filter2sql($node, $type);
-        $node = $node->nextSibling;
-        my $geom2 = filter2sql($node, $type);
-        return "ST_$name($geom1, $geom2)";
-
-    } elsif ($name eq 'DWithin') {
-        $node = $node->firstChild;
-        my $geom1 = filter2sql($node, $type);
-        $node = $node->nextSibling;
-        my $geom2 = filter2sql($node, $type);
-        $node = $node->nextSibling;
-        my $dist = filter2sql($node, $type);
-        return "ST_$name($geom1, $geom2, $dist)";
-
-    # temporal filter missing
-
-    # functions missing
-
-    } elsif ($name eq 'Filter') {
-        my %id = ( ResourceId => 1,
-                   FeatureId => 1,
-                   GmlObjectId => 1 );
-        $node = $node->firstChild;
-        my ($ns2, $name2) = parse_tag($node);
-        if ($id{$name2}) {
-            my $where = '';
-            for (my $id = $node; $id; $id = $id->nextSibling) {
-                $where .= filter2sql($id, $type).' OR ';
-            }
-            $where =~ s/ OR $//;
-            return $where;
-        } else {
-            return filter2sql($node, $type);
-        }
-    }
-}
-
-sub node2sql {
-    my ($node) = @_;
-    my $srs = get_integer($node->getAttribute('srsName')) // '';
-    $srs = ",$srs" if $srs ne '';
-    my ($ns, $name) = parse_tag($node);
-    my $wkt;
-
-    # should use Geo::OGC::Geometry
-
-    if ($name eq 'Box') {
-        my $env = strip($node->firstChild->textContent);
-        $env =~ s/ /,/;
-        return "ST_MakeEnvelope($env$srs)";
-    } elsif ($name eq 'Envelope') {
-        $node = $node->firstChild;
-        my $lc = strip($node->textContent); # gml:lowerCorner
-        $lc =~ s/ /,/;
-        $node = $node->nextSibling;
-        my $uc = strip($node->textContent); # gml:upperCorner
-        $uc =~ s/ /,/;
-        return "ST_MakeEnvelope($lc,$uc$srs)";
-    } elsif ($name eq 'Point') {
-        my $pos = $node->firstChild->firstChild->data;
-        $pos =~ s/,/ /;
-        $wkt = "POINT ($pos)";
-    } elsif ($name eq 'MultiPoint') { # not ok
-        my $pos = $node->firstChild->firstChild->data;
-        $pos =~ s/,/ /;
-        $wkt = "MULTIPOINT ($pos)";
-    } elsif ($name eq 'LineString') {
-        my @tmp = split /\s+/, $node->firstChild->firstChild->data;
-        my @pos;
-        for (my $i = 0; $i < @tmp; $i+=2) {
-            push @pos, $tmp[$i].' '.$tmp[$i+1];
-        }
-        $wkt = "LINESTRING (".join(', ',@pos).")";
-    } elsif ($name eq 'MultiLineString') { # not ok
-        my @tmp = split / /, $node->firstChild->firstChild->data;
-        my @pos;
-        for (my $i = 0; $i < @tmp; $i+=2) {
-            push @pos, $tmp[$i].' '.$tmp[$i+1];
-        }
-        $wkt = "MULTILINESTRING (".join(', ',@pos).")";
-    } elsif ($name eq 'Polygon') {
-        # Polygon.exterior.LinearRing.posList
-        my @tmp = split /\s+/, $node->firstChild->firstChild->firstChild->textContent;
-        my @pos;
-        for (my $i = 0; $i < @tmp; $i+=2) {
-            push @pos, strip($tmp[$i]).' '.strip($tmp[$i+1]);
-        }
-        $wkt = "POLYGON ((".join(', ',@pos)."))";
-    } elsif ($name eq 'MultiPolygon') { # not ok
-        my @tmp = split / /, $node->firstChild->firstChild->firstChild->data;
-        my @pos;
-        for (my $i = 0; $i < @tmp; $i+=2) {
-            push @pos, $tmp[$i].' '.$tmp[$i+1];
-        }
-        $wkt = "MULTIPOLYGON ((".join(', ',@pos)."))";
-    } elsif ($name eq 'MultiGeometry') { # not ok
-        my @tmp = split / /, $node->firstChild->firstChild->firstChild->data;
-        my @pos;
-        for (my $i = 0; $i < @tmp; $i+=2) {
-            push @pos, $tmp[$i].' '.$tmp[$i+1];
-        }
-        $wkt = "MULTIGEOMETRY ((".join(', ',@pos)."))";
-    }
-    
-    return "ST_GeometryFromText('$wkt'$srs)";
-}
-
-sub error {
-    my ($self, $msg) = @_;
-    if (!$msg->{debug}) {
-        Geo::OGC::Service::error($self->{responder}, $msg);
-    } else {
-        my $json = JSON->new;
-        $json->allow_blessed([1]);
-        my $writer = $self->{responder}->([200, [ 'Content-Type' => 'application/json',
-                                                  'Content-Encoding' => 'UTF-8' ]]);
-        $writer->write($json->encode($msg->{debug}));
-        $writer->close;
-    }
-}
-
-sub parse_tag {
-    my $node = shift;
-    return parse_name($node->nodeName);
-}
-
-sub parse_name {
-    my $name = shift;
-    my $ns = '';
-    ($ns, $name) = split /:/, $name if $name =~ /:/;
-    return ($ns, $name);
-}
-
-sub log {
-    my ($self, $msg) = @_;
-    say STDERR Dumper($msg);
-}
-
 sub DataSource2dbi {
     my $DataSource = shift;
     my $dbi = 'dbi:'.$DataSource;
@@ -1653,49 +1278,12 @@ sub DataSource2dbi {
     return $dbi;
 }
 
-sub latest_version {
-    my $versions = shift;
-    return undef unless defined $versions;
-    my @versions = split /\s*,\s*/, $versions;
-    for (@versions) {
-        my ($a,$b,$c) = split /\./;
-        $_ = $a*10000+$b*100+$c;
-    }
-    @versions = sort {$b <=> $a} @versions;
-    my ($a,$b,$c) = $versions[0] =~ /^(\d+)(\d\d)(\d\d)$/;
-    for ($a,$b,$c) {
-        $_ = int($_);
-    }
-    return "$a.$b.$c";
-}
-
-sub strip {
-    my $s = shift;
-    $s =~ s/^\s+//;
-    $s =~ s/\s+$//;
-    return $s;
-}
-    
 sub fallback {
     for my $arg (@_) {
         return $arg if defined $arg;
     }
 }
 
-sub get_integer {
-    my $s = shift;
-    if (not defined $s) {
-        if (@_) {
-            return get_integer(shift);
-        } else {
-            return undef;
-        }
-    }
-    if ($s =~ /(\d+)/) {
-        return $1;
-    }
-    return undef;
-}
 sub split_to_listref {
     my $s = shift;
     return undef unless $s;
